import logging
import pathlib
import re
from datetime import datetime
from ftplib import FTP
from itertools import chain
from multiprocessing.pool import ThreadPool
from typing import Any, List, Optional, Tuple

from numpy import append, arange
from tqdm import tqdm

from copernicus_marine_client.catalogue_parser.request_structure import (
    GetRequest,
)
from copernicus_marine_client.core_functions.utils import (
    flatten,
    get_unique_filename,
)
from copernicus_marine_client.download_functions.download_get import (
    download_get,
)

# /////////////////////////////
# ---Using ftplib
# /////////////////////////////

logger = logging.getLogger("copernicus_marine_root_logger")


def download_ftp(
    username: str,
    password: str,
    get_request: GetRequest,
) -> list[pathlib.Path]:
    filenames_in, filenames_out, host = download_get(
        username,
        password,
        get_request,
        download_header,
        create_filenames_out,
    )
    return download_files(
        host, username, password, filenames_in, filenames_out
    )


def download_files(
    host: str,
    username: str,
    password: str,
    filenames_in: List[str],
    filenames_out: List[pathlib.Path],
) -> list[pathlib.Path]:
    pool = ThreadPool()
    nfiles_per_process, nfiles = 1, len(filenames_in)
    indexes = append(
        arange(0, nfiles, nfiles_per_process, dtype=int),
        nfiles,
    )
    groups_in_files = [
        filenames_in[indexes[i] : indexes[i + 1]]
        for i in range(len(indexes) - 1)
    ]
    groups_out_files = [
        filenames_out[indexes[i] : indexes[i + 1]]
        for i in range(len(indexes) - 1)
    ]
    groups_in_files_count = len(groups_in_files)
    download_summary_list = pool.imap(
        _download_files,
        zip(
            [host] * groups_in_files_count,
            [username] * groups_in_files_count,
            [password] * groups_in_files_count,
            groups_in_files,
            groups_out_files,
        ),
    )
    download_summary = list(
        tqdm(download_summary_list, total=len(groups_in_files))
    )
    return flatten(download_summary)


def download_header(
    data_path: str, regex: Optional[str], username: str, password: str
) -> Tuple[str, str, list[str], float]:
    (host, path) = parse_ftp_dataset_url(data_path)
    message = "You requested the download of the following files:\n"
    total_size = 0
    with FTP(host) as ftp:
        ftp.login(user=username, passwd=password)
        raw_filenames = get_filenames_recursively(ftp, path)
        if regex:
            regex_compiled = re.compile(regex)
            filenames = list(filter(regex_compiled.match, raw_filenames))
        else:
            filenames = raw_filenames
    pool = ThreadPool()
    nfilenames_per_process, nfilenames = 100, len(filenames)
    indexes = append(
        arange(0, nfilenames, nfilenames_per_process, dtype=int),
        nfilenames,
    )
    groups_filenames = [
        filenames[indexes[i] : indexes[i + 1]] for i in range(len(indexes) - 1)
    ]
    results_size = pool.map(
        get_filename_size_tuple,
        zip(
            [host] * len(groups_filenames),
            [username] * len(groups_filenames),
            [password] * len(groups_filenames),
            groups_filenames,
        ),
    )
    flattened_results_size = list(chain(*results_size))

    total_size += sum([int(res[1]) for res in flattened_results_size])
    for result in flattened_results_size[:20]:
        message += str(result[0])
        message += f" - {format_file_size(float(result[1]))}"
        message += f" - {result[2]}\n"

    if len(flattened_results_size) > 20:
        message += f"Printed 20 out of {len(flattened_results_size)} files\n"
    message += (
        f"\nTotal size of the download: {format_file_size(total_size)}\n\n"
    )
    return (message, host, filenames, total_size)


def get_filenames_recursively(
    ftp: FTP, path: str, extensions: list[str] = [".nc"]
) -> list[str]:
    if any(extension in path for extension in extensions):
        # path is a file
        return [path]
    elif len(ftp.nlst(path)) == 0:
        # empty dir
        return []
    elif ftp.nlst(path)[0] == path:
        # path is a file
        return [path]
    else:
        # path is a dir
        return [
            filename
            for element in ftp.nlst(path)
            for filename in get_filenames_recursively(ftp, element)
        ]


def get_filename_size_tuple(
    tuple_ftp_filename: Tuple[str, str, str, list[str]]
) -> list[Tuple[str, Any, str]]:
    host, username, password, filenames = tuple_ftp_filename
    with FTP(host) as ftp:
        ftp.login(user=username, passwd=password)
        list_tuples = [
            (
                filename,
                ftp.size(filename),
                get_last_modified_datetime(ftp, filename),
            )
            for filename in filenames
        ]
    return list_tuples


def get_last_modified_datetime(ftp: FTP, filename: str) -> str:
    result_code_and_datetime = ftp.voidcmd("MDTM " + filename)
    result_code = result_code_and_datetime.split(" ")[0]
    if result_code != "213":
        return "N/C"
    ftp_datetime = result_code_and_datetime.split(" ")[1]
    datetime_iso = datetime.strptime(ftp_datetime, "%Y%m%d%H%M%S").strftime(
        "%Y-%m-%dT%H:%M:%S:%SZ"
    )
    return datetime_iso


def _download_files(
    tuple_ftp_filename: Tuple[str, str, str, list[str], list[pathlib.Path]],
) -> list[pathlib.Path]:
    def _ftp_file_download(ftp: FTP, file_in: str, file_out: pathlib.Path):
        """
        Download ONE file and return a string of the result
        """
        pathlib.Path.mkdir(file_out.parent, parents=True, exist_ok=True)
        with open(file_out, "wb") as fp:

            def callback(data):
                fp.write(data)

            res = ftp.retrbinary(f"RETR {file_in}", callback)

            if not res.startswith("226 Transfer complete"):
                logger.error(f"Download {file_in} failed")
                if file_out.is_file():
                    file_out.unlink()
                summary_string = file_out
            else:
                summary_string = file_out
        return summary_string

    host, username, password, filenames_in, filenames_out = tuple_ftp_filename
    download_summary: list[pathlib.Path] = []
    with FTP(host) as ftp:
        ftp.login(user=username, passwd=password)
        for file_in, file_out in zip(filenames_in, filenames_out):
            download_summary.append(_ftp_file_download(ftp, file_in, file_out))
    return download_summary


# /////////////////////////////
# --- Tools
# /////////////////////////////


def parse_ftp_dataset_url(data_path: str) -> tuple[str, str]:
    host = data_path[len("ftp://") :].split("/")[0]
    path = data_path[len("ftp://" + host + "/") :]
    return (host, path)


def create_filenames_out(
    filenames_in: list[str],
    overwrite: bool,
    output_directory: pathlib.Path = pathlib.Path("."),
    no_directories=False,
) -> list[pathlib.Path]:
    filenames_out = []
    for filename_in in filenames_in:
        filename_out = output_directory
        if no_directories:
            filename_out = (
                pathlib.Path(filename_out) / pathlib.Path(filename_in).name
            )
        elif filename_in.startswith("Core/"):
            filename_out = filename_out / pathlib.Path(
                filename_in[len("Core/") :]
            )

        filename_out = get_unique_filename(
            filepath=filename_out, overwrite_option=overwrite
        )

        filenames_out.append(filename_out)
    return filenames_out


def format_file_size(
    size: float, decimals: int = 2, binary_system: bool = False
) -> str:
    if binary_system:
        units: list[str] = [
            "B",
            "KiB",
            "MiB",
            "GiB",
            "TiB",
            "PiB",
            "EiB",
            "ZiB",
        ]
        largest_unit: str = "YiB"
        step: int = 1024
    else:
        units = ["B", "kB", "MB", "GB", "TB", "PB", "EB", "ZB"]
        largest_unit = "YB"
        step = 1000

    for unit in units:
        if size < step:
            return ("%." + str(decimals) + "f %s") % (size, unit)
        size /= step

    return ("%." + str(decimals) + "f %s") % (size, largest_unit)



import json
from typing import Any

from copernicus_marine_client.core_functions.describe import describe_function
from copernicus_marine_client.python_interface.exception_handler import (
    log_exception_and_exit,
)


@log_exception_and_exit
def describe(
    include_description: bool = False,
    include_datasets: bool = False,
    include_keywords: bool = False,
    contains: list[str] = [],
    overwrite_metadata_cache: bool = False,
    no_metadata_cache: bool = False,
    staging: bool = False,
) -> dict[str, Any]:
    """
    Retrieve metadata information from the Copernicus Marine catalogue.

    This function fetches metadata information from the Copernicus Marine catalogue
    based on specified parameters and options.

    Args:
        include_description (bool, optional): Whether to include description for each item. Defaults to False.
        include_datasets (bool, optional): Whether to include dataset information. Defaults to False.
        include_keywords (bool, optional): Whether to include keywords for each item. Defaults to False.
        contains (list[str], optional): List of strings to filter items containing these values. Defaults to [].
        overwrite_metadata_cache (bool, optional): Whether to overwrite the metadata cache. Defaults to False.
        no_metadata_cache (bool, optional): Whether to skip using the metadata cache. Defaults to False.

    Returns:
        dict[str, Any]: A dictionary containing the retrieved metadata information.
    """  # noqa
    catalogue_json = describe_function(
        include_description,
        include_datasets,
        include_keywords,
        contains,
        overwrite_metadata_cache,
        no_metadata_cache,
        staging=staging,
    )
    catalogue = json.loads(catalogue_json)
    return catalogue



import logging
import pathlib
from datetime import datetime
from importlib.metadata import version
from typing import Callable, Iterable, Iterator, Optional, TypeVar

import xarray
from requests import PreparedRequest

logger = logging.getLogger("copernicus_marine_root_logger")

OVERWRITE_SHORT_OPTION = "--overwrite"
OVERWRITE_LONG_OPTION = "--overwrite-output-data"
OVERWRITE_OPTION_HELP_TEXT = (
    "If specified and if the file already exists on destination, then it will be "
    "overwritten instead of creating new one with unique index."
)

FORCE_DOWNLOAD_CLI_PROMPT_MESSAGE = "Do you want to proceed with download?"

DEFAULT_CLIENT_BASE_DIRECTORY = (
    pathlib.Path.home() / ".copernicus-marine-client"
)

# Client name, to be used in custom query parameters for each request to MDS S3
# endpoints, in order to track statistics of usage
CLIENT_NAME = "copernicus-marine-client"

DATETIME_SUPPORTED_FORMATS = [
    "%Y",
    "%Y-%m-%d",
    "%Y-%m-%dT%H:%M:%S",
    "%Y-%m-%d %H:%M:%S",
    "%Y-%m-%dT%H:%M:%S.%fZ",
]


def get_unique_filename(
    filepath: pathlib.Path, overwrite_option: bool
) -> pathlib.Path:
    if not overwrite_option:
        parent = filepath.parent
        filename = filepath.stem
        extension = filepath.suffix
        counter = 1

        while filepath.exists():
            filepath = parent / (
                filename + "_(" + str(counter) + ")" + extension
            )
            counter += 1

    return filepath


_T = TypeVar("_T")
_S = TypeVar("_S")


def map_reject_none(
    function: Callable[[_S], Optional[_T]], iterable: Iterable[_S]
) -> Iterable[_T]:
    return (element for element in map(function, iterable) if element)


def next_or_raise_exception(
    iterator: Iterator[_T], exception_to_raise: Exception
) -> _T:
    try:
        return next(iterator)
    except Exception as exception:
        raise exception_to_raise from exception


def flatten(list: list[list[_T]]) -> list[_T]:
    return [item for sublist in list for item in sublist]


def construct_url_with_query_params(url, query_params: dict) -> Optional[str]:
    req = PreparedRequest()
    req.prepare_url(url, query_params)
    return req.url


def construct_query_params_for_marine_data_store_monitoring(
    username: Optional[str] = None,
) -> dict:
    query_params = {"x-cop-client": CLIENT_NAME}
    if username:
        query_params["x-cop-user"] = username
    return query_params


class WrongDatetimeFormat(Exception):
    ...


def datetime_parser(string: str):
    for format in DATETIME_SUPPORTED_FORMATS:
        try:
            return datetime.strptime(string, format)
        except ValueError:
            pass
    raise WrongDatetimeFormat(string)


def add_copernicus_marine_client_version_in_dataset_attributes(
    dataset: xarray.Dataset,
) -> xarray.Dataset:
    dataset.attrs["copernicus_marine_client_version"] = version(
        "copernicus-marine-client"
    )
    return dataset



import logging
from dataclasses import dataclass
from enum import Enum
from itertools import chain
from typing import List, Literal, Optional, Tuple

import xarray

from copernicus_marine_client.catalogue_parser.catalogue_parser import (
    CopernicusMarineCatalogue,
    CopernicusMarineDatasetService,
    CopernicusMarineDatasetServiceType,
    CopernicusMarineProductDataset,
    get_dataset_and_suffix_path_from_url,
    get_dataset_from_id,
)
from copernicus_marine_client.catalogue_parser.request_structure import (
    DatasetTimeAndGeographicalSubset,
)
from copernicus_marine_client.core_functions.utils import (
    construct_query_params_for_marine_data_store_monitoring,
    next_or_raise_exception,
)
from copernicus_marine_client.download_functions.subset_parameters import (
    DepthParameters,
    GeographicalParameters,
    LatitudeParameters,
    LongitudeParameters,
    TemporalParameters,
)
from copernicus_marine_client.download_functions.subset_xarray import subset

logger = logging.getLogger("copernicus_marine_root_logger")


class _Command(Enum):
    GET = "get"
    SUBSET = "subset"
    LOAD = "load"


@dataclass(frozen=True)
class Command:
    command_name: _Command
    service_types_by_priority: List[CopernicusMarineDatasetServiceType]

    def service_names(self) -> List[str]:
        return list(
            map(
                lambda service_type: service_type.service_name.value,
                self.service_types_by_priority,
            )
        )

    def service_short_names(self) -> List[str]:
        return list(
            map(
                lambda service_type: service_type.short_name.value,
                self.service_types_by_priority,
            )
        )

    def service_aliases(self) -> List[str]:
        return list(
            chain(
                *map(
                    lambda service_type: service_type.aliases(),
                    self.service_types_by_priority,
                )
            )
        )


class CommandType(Command, Enum):
    SUBSET = _Command.SUBSET, [
        CopernicusMarineDatasetServiceType.GEOSERIES,
        CopernicusMarineDatasetServiceType.TIMESERIES,
        CopernicusMarineDatasetServiceType.OPENDAP,
        CopernicusMarineDatasetServiceType.MOTU,
    ]
    GET = _Command.GET, [
        CopernicusMarineDatasetServiceType.FILES,
        CopernicusMarineDatasetServiceType.FTP,
    ]
    LOAD = _Command.LOAD, [
        CopernicusMarineDatasetServiceType.GEOSERIES,
        CopernicusMarineDatasetServiceType.TIMESERIES,
        CopernicusMarineDatasetServiceType.OPENDAP,
    ]


def assert_service_type_for_command(
    service_type: CopernicusMarineDatasetServiceType, command_type: CommandType
) -> CopernicusMarineDatasetServiceType:
    return next_or_raise_exception(
        (
            service_type
            for service_type in command_type.service_types_by_priority
        ),
        service_type_does_not_exist_for_command(service_type, command_type),
    )


class ServiceDoesNotExistForCommand(Exception):
    def __init__(self, service_name, command_name, available_services):
        super().__init__()
        self.__setattr__(
            "custom_exception_message",
            f"Service {service_name} "
            f"does not exist for command {command_name}. "
            f"Possible service{'s' if len(available_services) > 1 else ''}: "
            f"{available_services}",
        )


def service_type_does_not_exist_for_command(
    service_type: CopernicusMarineDatasetServiceType, command_type: CommandType
) -> ServiceDoesNotExistForCommand:
    return service_does_not_exist_for_command(
        service_type.service_name.value, command_type
    )


def service_does_not_exist_for_command(
    service_name: str, command_type: CommandType
) -> ServiceDoesNotExistForCommand:
    return ServiceDoesNotExistForCommand(
        service_name,
        command_type.command_name.value,
        command_type.service_aliases(),
    )


def _select_forced_service(
    dataset: CopernicusMarineProductDataset,
    force_service_type: CopernicusMarineDatasetServiceType,
    command_type: CommandType,
) -> CopernicusMarineDatasetService:
    logger.info(
        f"You forced selection of service: "
        f"{force_service_type.service_name.value}"
    )
    return next_or_raise_exception(
        (
            service
            for service in dataset.services
            if service.service_type == force_service_type
        ),
        service_not_available_error(dataset, command_type),
    )


def _get_best_arco_service_type(
    dataset_subset: DatasetTimeAndGeographicalSubset,
    dataset_url: str,
) -> Literal[
    CopernicusMarineDatasetServiceType.TIMESERIES,
    CopernicusMarineDatasetServiceType.GEOSERIES,
]:
    dataset = xarray.open_zarr(
        dataset_url,
        # Pass custom query parameters for MDS's Monitoring
        # (note that here we don't have 'username', shall we pass it?)
        storage_options={
            "params": construct_query_params_for_marine_data_store_monitoring(),
        },
    )
    subset_dataset_coordinates = subset(
        data_object=dataset.coords,
        variables=None,
        geographical_parameters=GeographicalParameters(
            latitude_parameters=LatitudeParameters(
                minimum_latitude=dataset_subset.minimum_latitude,
                maximum_latitude=dataset_subset.maximum_latitude,
            ),
            longitude_parameters=LongitudeParameters(
                minimum_longitude=dataset_subset.minimum_longitude,
                maximum_longitude=dataset_subset.maximum_longitude,
            ),
        ),
        temporal_parameters=TemporalParameters(
            start_datetime=dataset_subset.start_datetime,
            end_datetime=dataset_subset.end_datetime,
        ),
        depth_parameters=DepthParameters(
            minimum_depth=None, maximum_depth=None
        ),
    )
    dataset_coordinates = dataset.coords

    geographical_dimensions = (
        dataset_coordinates["latitude"].size
        * dataset_coordinates["longitude"].size
    )
    subset_geographical_dimensions = (
        subset_dataset_coordinates["latitude"].size
        * subset_dataset_coordinates["longitude"].size
    )
    temporal_dimensions = dataset_coordinates["time"].size
    subset_temporal_dimensions = subset_dataset_coordinates["time"].size

    geographical_coverage = (
        subset_geographical_dimensions / geographical_dimensions
    )
    temporal_coverage = subset_temporal_dimensions / temporal_dimensions

    if geographical_coverage >= temporal_coverage:
        return CopernicusMarineDatasetServiceType.GEOSERIES
    return CopernicusMarineDatasetServiceType.TIMESERIES


def _get_first_available_service_type(
    dataset: CopernicusMarineProductDataset,
    command_type: CommandType,
    dataset_available_service_types: list[CopernicusMarineDatasetServiceType],
) -> CopernicusMarineDatasetServiceType:
    return next_or_raise_exception(
        (
            service_type
            for service_type in command_type.service_types_by_priority
            if service_type in dataset_available_service_types
        ),
        no_service_available_for_command(dataset.dataset_id, command_type),
    )


def _select_service_by_priority(
    dataset: CopernicusMarineProductDataset,
    command_type: CommandType,
    dataset_subset: Optional[DatasetTimeAndGeographicalSubset],
) -> CopernicusMarineDatasetService:
    dataset_available_service_types = [
        service.service_type for service in dataset.services
    ]
    first_available_service_type = _get_first_available_service_type(
        dataset, command_type, dataset_available_service_types
    )
    first_available_service = dataset.get_service_by_service_type(
        first_available_service_type
    )
    if (
        CopernicusMarineDatasetServiceType.GEOSERIES
        in dataset_available_service_types
        and CopernicusMarineDatasetServiceType.TIMESERIES
        in dataset_available_service_types
        and command_type in [CommandType.SUBSET, CommandType.LOAD]
        and dataset_subset is not None
    ):
        best_arco_service_type: CopernicusMarineDatasetServiceType = (
            _get_best_arco_service_type(
                dataset_subset, first_available_service.uri
            )
        )
        return dataset.get_service_by_service_type(best_arco_service_type)
    else:
        return first_available_service


def _select_service_from_url(
    dataset: CopernicusMarineProductDataset,
    dataset_url: str,
    command_type: CommandType,
) -> CopernicusMarineDatasetService:
    url_service_type: CopernicusMarineDatasetServiceType = (
        _get_service_type_from_url(dataset_url)
    )
    assert_service_type_for_command(url_service_type, command_type)
    return dataset.get_service_by_service_type(url_service_type)


def _select_dataset_service(
    dataset: CopernicusMarineProductDataset,
    dataset_url: Optional[str],
    force_service: Optional[str],
    command_type: CommandType,
    dataset_subset: Optional[DatasetTimeAndGeographicalSubset] = None,
) -> CopernicusMarineDatasetService:
    if force_service:
        force_service_type = service_type_from_string(
            force_service, command_type
        )
        return _select_forced_service(
            dataset, force_service_type, command_type
        )
    if dataset_url:
        return _select_service_from_url(dataset, dataset_url, command_type)
    return _select_service_by_priority(dataset, command_type, dataset_subset)


def get_dataset_service_and_suffix_path(
    catalogue: CopernicusMarineCatalogue,
    dataset_id: Optional[str],
    dataset_url: Optional[str],
    force_service_type: Optional[str],
    command_type: CommandType,
    dataset_subset: Optional[DatasetTimeAndGeographicalSubset] = None,
) -> Tuple[CopernicusMarineDatasetService, str, Optional[str]]:
    if dataset_id is None and dataset_url is None:
        syntax_error = SyntaxError(
            "Must specify at least one of "
            "'dataset_url' or 'dataset_id' options"
        )
        raise syntax_error
    if dataset_id is not None and dataset_url is not None:
        syntax_error = SyntaxError(
            "Must specify only one of 'dataset_url' or 'dataset_id' options"
        )
        raise syntax_error
    if dataset_id:
        dataset = get_dataset_from_id(
            catalogue=catalogue, dataset_id=dataset_id
        )
        return (
            _select_dataset_service(
                dataset,
                None,
                force_service_type,
                command_type,
                dataset_subset,
            ),
            dataset_id,
            None,
        )
    if dataset_url:
        dataset, suffix_path = get_dataset_and_suffix_path_from_url(
            catalogue=catalogue, dataset_url=dataset_url
        )
        return (
            _select_dataset_service(
                dataset, dataset_url, force_service_type, command_type
            ),
            dataset.dataset_id,
            suffix_path,
        )
    raise NotPossibleError()


class NotPossibleError(Exception):
    ...


def _get_service_type_from_url(
    dataset_url,
) -> CopernicusMarineDatasetServiceType:
    if dataset_url.startswith("ftp://"):
        service_type = CopernicusMarineDatasetServiceType.FTP
    elif "/motu-web/Motu" in dataset_url:
        service_type = CopernicusMarineDatasetServiceType.MOTU
    elif "/thredds/dodsC/" in dataset_url:
        service_type = CopernicusMarineDatasetServiceType.OPENDAP
    elif "/mdl-arco-time" in dataset_url:
        service_type = CopernicusMarineDatasetServiceType.TIMESERIES
    elif "/mdl-arco-geo" in dataset_url:
        service_type = CopernicusMarineDatasetServiceType.GEOSERIES
    elif "/mdl-native" in dataset_url:
        service_type = CopernicusMarineDatasetServiceType.FILES
    else:
        exception = ValueError(f"No service matching url: {dataset_url}")
        raise exception
    return service_type


class ServiceNotAvailable(Exception):
    ...


def service_not_available_error(
    dataset: CopernicusMarineProductDataset, command_type: CommandType
) -> ServiceNotAvailable:
    dataset_available_service_types = [
        service.service_type.short_name.value
        for service in dataset.services
        if service.service_type in command_type.service_types_by_priority
    ]
    return ServiceNotAvailable(
        f"Available services for dataset {dataset.dataset_id}: "
        f"{dataset_available_service_types}"
    )


class NoServiceAvailable(Exception):
    ...


def no_service_available_for_command(
    dataset_id: str, command_type: CommandType
) -> NoServiceAvailable:
    return NoServiceAvailable(
        f"No service available for dataset {dataset_id} "
        f"with command {command_type.command_name.value}"
    )


def service_type_from_string(
    string: str, command_type: CommandType
) -> CopernicusMarineDatasetServiceType:
    return next_or_raise_exception(
        (
            service_type
            for service_type in command_type.service_types_by_priority
            if string in service_type.aliases()
        ),
        service_does_not_exist_for_command(string, command_type),
    )



import pathlib
from datetime import datetime
from typing import List, Optional, Union

from copernicus_marine_client.core_functions.deprecated import (
    deprecated_python_option,
)
from copernicus_marine_client.core_functions.deprecated_options import (
    DEPRECATED_OPTIONS,
)
from copernicus_marine_client.core_functions.subset import subset_function
from copernicus_marine_client.python_interface.exception_handler import (
    log_exception_and_exit,
)
from copernicus_marine_client.python_interface.utils import homogenize_datetime


@deprecated_python_option(**DEPRECATED_OPTIONS.dict_old_names_to_new_names)
@log_exception_and_exit
def subset(
    dataset_url: Optional[str] = None,
    dataset_id: Optional[str] = None,
    username: Optional[str] = None,
    password: Optional[str] = None,
    variables: Optional[List[str]] = None,
    minimum_longitude: Optional[float] = None,
    maximum_longitude: Optional[float] = None,
    minimum_latitude: Optional[float] = None,
    maximum_latitude: Optional[float] = None,
    minimum_depth: Optional[float] = None,
    maximum_depth: Optional[float] = None,
    vertical_dimension_as_originally_produced: bool = True,
    start_datetime: Optional[Union[datetime, str]] = None,
    end_datetime: Optional[Union[datetime, str]] = None,
    output_filename: Optional[str] = None,
    force_service: Optional[str] = None,
    request_file: Optional[Union[pathlib.Path, str]] = None,
    output_directory: Optional[Union[pathlib.Path, str]] = None,
    credentials_file: Optional[Union[pathlib.Path, str]] = None,
    motu_api_request: Optional[str] = None,
    force_download: bool = False,
    overwrite_output_data: bool = False,
    overwrite_metadata_cache: bool = False,
    no_metadata_cache: bool = False,
    staging: bool = False,
) -> pathlib.Path:
    """
    Extracts a subset of data from a specified dataset using given parameters.

    Args:
        dataset_url (str, optional): URL of the dataset source.
        dataset_id (str, optional): Identifier for the dataset.
        username (str, optional): Username for authentication.
        password (str, optional): Password for authentication.
        variables (List[str], optional): List of variable names to extract.
        minimum_longitude (float, optional): Minimum longitude value for spatial subset.
        maximum_longitude (float, optional): Maximum longitude value for spatial subset.
        minimum_latitude (float, optional): Minimum latitude value for spatial subset.
        maximum_latitude (float, optional): Maximum latitude value for spatial subset.
        minimum_depth (float, optional): Minimum depth value for vertical subset.
        maximum_depth (float, optional): Maximum depth value for vertical subset.
        vertical_dimension_as_originally_produced (bool, optional): Use original vertical dimension.
        start_datetime (datetime, optional): Start datetime for temporal subset.
        end_datetime (datetime, optional): End datetime for temporal subset.
        output_filename (str, optional): Output filename for the subsetted data.
        force_service (str, optional): Force use of specified data service.
        request_file (Union[pathlib.Path, str], optional): Path to request file.
        output_directory (Union[pathlib.Path, str], optional): Directory to save output files.
        credentials_file (Union[pathlib.Path, str], optional): Path to credentials file.
        motu_api_request (str, optional): MOTU API request string.
        force_download (bool, optional): Skip confirmation before download.
        overwrite_output_data (bool, optional): Overwrite existing output data if True.
        overwrite_metadata_cache (bool, optional): Overwrite existing metadata cache if True.
        no_metadata_cache (bool, optional): Disable metadata caching if True.

    Returns:
        pathlib.Path: Path to the generated subsetted data file.
    """  # noqa
    request_file = pathlib.Path(request_file) if request_file else None
    output_directory = (
        pathlib.Path(output_directory) if output_directory else None
    )
    credentials_file = (
        pathlib.Path(credentials_file) if credentials_file else None
    )

    start_datetime = homogenize_datetime(start_datetime)
    end_datetime = homogenize_datetime(end_datetime)

    return subset_function(
        dataset_url,
        dataset_id,
        username,
        password,
        variables,
        minimum_longitude,
        maximum_longitude,
        minimum_latitude,
        maximum_latitude,
        minimum_depth,
        maximum_depth,
        vertical_dimension_as_originally_produced,
        start_datetime,
        end_datetime,
        output_filename,
        force_service,
        request_file,
        output_directory,
        credentials_file,
        motu_api_request,
        force_download,
        overwrite_output_data,
        overwrite_metadata_cache,
        no_metadata_cache,
        staging=staging,
    )



import logging
from datetime import datetime
from decimal import Decimal
from typing import List, Literal, Optional, Union

import numpy as np
import xarray

from copernicus_marine_client.download_functions.subset_parameters import (
    DepthParameters,
    GeographicalParameters,
    LatitudeParameters,
    LongitudeParameters,
    TemporalParameters,
)

logger = logging.getLogger("copernicus_marine_root_logger")

COORDINATES_LABEL = {
    "latitude": ["latitude", "nav_lat", "x", "lat"],
    "longitude": ["longitude", "nav_lon", "y", "lon"],
    "time": ["time_counter", "time"],
    "depth": ["depth", "deptht", "elevation"],
}


class MinimumLongitudeGreaterThanMaximumLongitude(Exception):
    pass


class VariableDoesNotExistInTheDataset(Exception):
    def __init__(self, variable):
        super().__init__()
        self.__setattr__(
            "custom_exception_message",
            f"The variable '{variable}' is neither a variable or a standard name in the dataset",  # noqa
        )


def _dataset_custom_sel(
    dataset: xarray.Dataset,
    coord_type: Literal["latitude", "longitude", "depth", "time"],
    coord_selection: Union[float, slice, datetime, None],
    method: Union[str, None] = None,
) -> xarray.Dataset:
    for coord_label in COORDINATES_LABEL[coord_type]:
        if coord_label in dataset.coords:
            tmp_dataset = dataset.sel(
                {coord_label: coord_selection}, method=method
            )
            if tmp_dataset.coords[coord_label].size == 0 and isinstance(
                coord_selection, slice
            ):
                dataset = dataset.sel(
                    {coord_label: coord_selection.start}, method="nearest"
                )
            else:
                dataset = tmp_dataset
    return dataset


def _coordinates_custom_sel(
    coordinates: xarray.Coordinates,
    coord_type: Literal["latitude", "longitude", "depth", "time"],
    coord_selection: Union[float, slice, datetime, None],
    method: Union[str, None] = None,
) -> xarray.Coordinates:
    for coord_label in COORDINATES_LABEL[coord_type]:
        if coord_label in coordinates:
            coordinates = coordinates.assign(
                coordinates[coord_label]
                .sel({coord_label: coord_selection}, method=method)
                .coords
            )
    return coordinates


def _xarray_data_object_custom_sel(
    data_object: Union[xarray.Dataset, xarray.Coordinates],
    coord_type: Literal["latitude", "longitude", "depth", "time"],
    coord_selection: Union[float, slice, datetime, None],
    method: Union[str, None] = None,
) -> Union[xarray.Dataset, xarray.Coordinates]:
    if isinstance(data_object, xarray.Dataset):
        return _dataset_custom_sel(
            data_object, coord_type, coord_selection, method
        )
    if isinstance(data_object, xarray.Coordinates):
        return _coordinates_custom_sel(
            data_object, coord_type, coord_selection, method
        )
    return data_object


def _update_dataset_attributes(dataset: xarray.Dataset):
    for coord_label in COORDINATES_LABEL["longitude"]:
        if coord_label in dataset.coords:
            attrs = dataset[coord_label].attrs
            if "valid_min" in attrs:
                attrs["valid_min"] += 180
            if "valid_max" in attrs:
                attrs["valid_max"] += 180
            dataset = dataset.assign_coords(
                {coord_label: dataset[coord_label] % 360}
            ).sortby(coord_label)
            dataset[coord_label].attrs = attrs
    return dataset


def _update_coordinates_attributes(coordinates: xarray.Coordinates):
    for coord_label in COORDINATES_LABEL["longitude"]:
        if coord_label in coordinates:
            attrs = coordinates[coord_label].attrs
            if "valid_min" in attrs:
                attrs["valid_min"] += 180
            if "valid_max" in attrs:
                attrs["valid_max"] += 180
            coordinates = coordinates.assign(
                coordinates[coord_label]
                .assign_coords({coord_label: coordinates[coord_label] % 360})
                .sortby(coord_label)
                .coords
            )
            coordinates[coord_label].attrs = attrs
    return coordinates


def _update_data_object_attributes(
    data_object: Union[xarray.Dataset, xarray.Coordinates]
):
    if isinstance(data_object, xarray.Dataset):
        return _update_dataset_attributes(data_object)
    if isinstance(data_object, xarray.Coordinates):
        return _update_coordinates_attributes(data_object)
    return data_object


def _latitude_subset(
    data_object: Union[xarray.Dataset, xarray.Coordinates],
    latitude_parameters: LatitudeParameters,
) -> xarray.Dataset:
    minimum_latitude = latitude_parameters.minimum_latitude
    maximum_latitude = latitude_parameters.maximum_latitude
    if minimum_latitude is not None or maximum_latitude is not None:
        latitude_selection = (
            minimum_latitude
            if minimum_latitude == maximum_latitude
            else slice(minimum_latitude, maximum_latitude)
        )
        latitude_method = (
            "nearest" if minimum_latitude == maximum_latitude else None
        )
        data_object = _xarray_data_object_custom_sel(
            data_object, "latitude", latitude_selection, latitude_method
        )
    return data_object


def _longitude_subset(
    data_object: Union[xarray.Dataset, xarray.Coordinates],
    longitude_parameters: LongitudeParameters,
) -> xarray.Dataset:
    minimum_longitude = longitude_parameters.minimum_longitude
    maximum_longitude = longitude_parameters.maximum_longitude
    if minimum_longitude is not None or maximum_longitude is not None:
        if minimum_longitude is not None and maximum_longitude is not None:
            if minimum_longitude > maximum_longitude:
                raise MinimumLongitudeGreaterThanMaximumLongitude(
                    "--minimum-longitude option must be smaller "
                    "or equal to --maximum-longitude"
                )
            if maximum_longitude - minimum_longitude >= 360:
                longitude_selection: Union[float, slice, None] = None
            elif minimum_longitude == maximum_longitude:
                longitude_selection = longitude_modulus(minimum_longitude)
                longitude_method = "nearest"
            else:
                minimum_longitude_modulus = longitude_modulus(
                    minimum_longitude
                )
                maximum_longitude_modulus = longitude_modulus(
                    maximum_longitude
                )
                if maximum_longitude_modulus < minimum_longitude_modulus:
                    maximum_longitude_modulus += 360
                    data_object = _update_data_object_attributes(data_object)
                longitude_selection = slice(
                    minimum_longitude_modulus,
                    maximum_longitude_modulus,
                )
                longitude_method = None
        else:
            longitude_selection = slice(minimum_longitude, maximum_longitude)
            longitude_method = None

        if longitude_selection is not None:
            data_object = _xarray_data_object_custom_sel(
                data_object, "longitude", longitude_selection, longitude_method
            )
    return data_object


def _temporal_subset(
    data_object: Union[xarray.Dataset, xarray.Coordinates],
    temporal_parameters: TemporalParameters,
) -> xarray.Dataset:
    start_datetime = temporal_parameters.start_datetime
    end_datetime = temporal_parameters.end_datetime
    if start_datetime is not None or end_datetime is not None:
        temporal_selection = (
            start_datetime
            if start_datetime == end_datetime
            else slice(start_datetime, end_datetime)
        )
        temporal_method = "nearest" if start_datetime == end_datetime else None
        data_object = _xarray_data_object_custom_sel(
            data_object, "time", temporal_selection, temporal_method
        )
    return data_object


def _depth_subset(
    data_object: Union[xarray.Dataset, xarray.Coordinates],
    depth_parameters: DepthParameters,
) -> xarray.Dataset:
    def convert_elevation_to_depth(dataset: xarray.Dataset):
        if "elevation" in dataset.coords:
            attrs = dataset["elevation"].attrs
            dataset = dataset.reindex(elevation=dataset.elevation[::-1])
            dataset["elevation"] = dataset.elevation * (-1)
            dataset = dataset.rename({"elevation": "depth"})
            dataset.depth.attrs = attrs
        return dataset

    if (
        depth_parameters.vertical_dimension_as_originally_produced
        and isinstance(data_object, xarray.Dataset)
    ):
        data_object = convert_elevation_to_depth(data_object)
    minimum_depth = depth_parameters.minimum_depth
    maximum_depth = depth_parameters.maximum_depth
    if minimum_depth is not None or maximum_depth is not None:
        coords = (
            data_object.coords
            if isinstance(data_object, xarray.Dataset)
            else data_object
        )
        if "elevation" in coords:
            minimum_depth = (
                minimum_depth * -1.0 if minimum_depth is not None else None
            )
            maximum_depth = (
                maximum_depth * -1.0 if maximum_depth is not None else None
            )
            minimum_depth, maximum_depth = maximum_depth, minimum_depth

        depth_selection = (
            minimum_depth
            if minimum_depth == maximum_depth
            else slice(minimum_depth, maximum_depth)
        )
        depth_method = "nearest" if minimum_depth == maximum_depth else None
        data_object = _xarray_data_object_custom_sel(
            data_object, "depth", depth_selection, depth_method
        )
    return data_object


def _get_variable_name_from_standard_name(
    dataset: xarray.Dataset, standard_name: str
) -> Optional[str]:
    for variable_name in dataset.variables:
        if (
            hasattr(dataset[variable_name], "standard_name")
            and dataset[variable_name].standard_name == standard_name
        ):
            return str(variable_name)
    return None


def _variables_subset(
    dataset: xarray.Dataset, variables: List[str]
) -> xarray.Dataset:
    dataset_variables_filter = []

    for variable in variables:
        if variable in dataset.variables:
            dataset_variables_filter.append(variable)
        else:
            variable_name_from_standard_name = (
                _get_variable_name_from_standard_name(dataset, variable)
            )
            if variable_name_from_standard_name is not None:
                dataset_variables_filter.append(
                    variable_name_from_standard_name
                )
            else:
                raise VariableDoesNotExistInTheDataset(variable)
    return dataset[np.array(dataset_variables_filter)]


def subset(
    data_object: Union[xarray.Dataset, xarray.Coordinates],
    variables: Optional[List[str]],
    geographical_parameters: GeographicalParameters,
    temporal_parameters: TemporalParameters,
    depth_parameters: DepthParameters,
) -> xarray.Dataset:
    if variables and isinstance(data_object, xarray.Dataset):
        data_object = _variables_subset(data_object, variables)

    data_object = _latitude_subset(
        data_object, geographical_parameters.latitude_parameters
    )
    data_object = _longitude_subset(
        data_object, geographical_parameters.longitude_parameters
    )

    data_object = _temporal_subset(data_object, temporal_parameters)

    data_object = _depth_subset(data_object, depth_parameters)

    return data_object


def longitude_modulus(longitude: float) -> float:
    """
    Returns the equivalent longitude between -180 and 180
    """
    # We are using Decimal to avoid issue with rounding
    modulus = float(Decimal(str(longitude + 180)) % 360)
    # Modulus with python return a negative value if the denominator is negative
    # To counteract that, we add 360 if the result is < 0
    modulus = modulus if modulus >= 0 else modulus + 360
    return modulus - 180



import asyncio
import logging
from dataclasses import dataclass
from enum import Enum
from importlib.metadata import version
from itertools import chain, groupby, repeat
from json import loads
from typing import (
    Any,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Tuple,
    TypeVar,
    Union,
)

import aiohttp
import nest_asyncio
import pystac
import requests
from cachier import cachier
from tqdm import tqdm

from copernicus_marine_client.command_line_interface.exception_handler import (
    log_exception_debug,
)
from copernicus_marine_client.core_functions.utils import (
    DEFAULT_CLIENT_BASE_DIRECTORY,
    construct_query_params_for_marine_data_store_monitoring,
    map_reject_none,
    next_or_raise_exception,
)

logger = logging.getLogger("copernicus_marine_root_logger")

_S = TypeVar("_S")
_T = TypeVar("_T")


class _ServiceName(str, Enum):
    MOTU = "motu"
    OPENDAP = "opendap"
    GEOSERIES = "arco-geo-series"
    TIMESERIES = "arco-time-series"
    FILES = "original-files"
    FTP = "ftp"
    WMS = "wms"
    WMTS = "wmts"


class _ServiceShortName(str, Enum):
    MOTU = "motu"
    OPENDAP = "opendap"
    GEOSERIES = "geoseries"
    TIMESERIES = "timeseries"
    FILES = "files"
    FTP = "ftp"
    WMS = "wms"
    WMTS = "wmts"


MARINE_DATA_STORE_STAC_BASE_URL = "https://stac.marine.copernicus.eu/metadata"
MARINE_DATA_STORE_STAC_ROOT_CATALOG_URL = (
    MARINE_DATA_STORE_STAC_BASE_URL + "/catalog.stac.json"
)
MARINE_DATA_STORE_STAC_BASE_URL_STAGING = (
    "https://stac-dta.marine.copernicus.eu/metadata"
)
MARINE_DATA_STORE_STAC_ROOT_CATALOG_URL_STAGING = (
    MARINE_DATA_STORE_STAC_BASE_URL_STAGING + "/catalog.stac.json"
)


@dataclass(frozen=True)
class _Service:
    service_name: _ServiceName
    short_name: _ServiceShortName

    def aliases(self) -> List[str]:
        return (
            [self.service_name.value, self.short_name.value]
            if self.short_name.value != self.service_name.value
            else [self.service_name.value]
        )

    def to_json_dict(self):
        return {
            "service_name": self.service_name.value,
            "short_name": self.short_name.value,
        }


class CopernicusMarineDatasetServiceType(_Service, Enum):
    MOTU = _ServiceName.MOTU, _ServiceShortName.MOTU
    OPENDAP = _ServiceName.OPENDAP, _ServiceShortName.OPENDAP
    GEOSERIES = _ServiceName.GEOSERIES, _ServiceShortName.GEOSERIES
    TIMESERIES = (
        _ServiceName.TIMESERIES,
        _ServiceShortName.TIMESERIES,
    )
    FILES = _ServiceName.FILES, _ServiceShortName.FILES
    FTP = _ServiceName.FTP, _ServiceShortName.FTP
    WMS = _ServiceName.WMS, _ServiceShortName.WMS
    WMTS = _ServiceName.WMS, _ServiceShortName.WMTS


def _service_type_from_web_api_string(
    name: str,
) -> CopernicusMarineDatasetServiceType:
    class WebApi(Enum):
        MOTU = "motu"
        OPENDAP = "opendap"
        GEOSERIES = "timeChunked"
        TIMESERIES = "geoChunked"
        FILES = "native"
        FTP = "ftp"
        WMS = "wms"
        WMTS = "wmts"

    web_api_mapping = {
        WebApi.MOTU: CopernicusMarineDatasetServiceType.MOTU,
        WebApi.OPENDAP: CopernicusMarineDatasetServiceType.OPENDAP,
        WebApi.GEOSERIES: CopernicusMarineDatasetServiceType.GEOSERIES,
        WebApi.TIMESERIES: CopernicusMarineDatasetServiceType.TIMESERIES,
        WebApi.FILES: CopernicusMarineDatasetServiceType.FILES,
        WebApi.FTP: CopernicusMarineDatasetServiceType.FTP,
        WebApi.WMS: CopernicusMarineDatasetServiceType.WMS,
        WebApi.WMTS: CopernicusMarineDatasetServiceType.WMTS,
    }

    return next_or_raise_exception(
        (
            service_type
            for service_web_api, service_type in web_api_mapping.items()
            if service_web_api.value == name
        ),
        ServiceNotHandled(name),
    )


class ServiceNotHandled(Exception):
    ...


@dataclass
class CopernicusMarineDatasetService:
    service_type: CopernicusMarineDatasetServiceType
    uri: str


@dataclass
class CopernicusMarineDatasetCoordinates:
    coordinates_id: str
    units: str
    minimum_value: Optional[float]
    maximum_value: Optional[float]
    step: Optional[float]
    values: Optional[str]


@dataclass
class CopernicusMarineDatasetVariable:
    short_name: str
    standard_name: str
    units: str
    bbox: Tuple[float, float, float, float]
    coordinates: list[CopernicusMarineDatasetCoordinates]


@dataclass
class CopernicusMarineProductDataset:
    dataset_id: str
    dataset_name: str
    services: list[CopernicusMarineDatasetService]
    variables: list[CopernicusMarineDatasetVariable]

    def get_available_service_types(
        self,
    ) -> list[CopernicusMarineDatasetServiceType]:
        return list(map(lambda service: service.service_type, self.services))

    def get_service_by_service_type(
        self, service_type: CopernicusMarineDatasetServiceType
    ):
        return next(
            service
            for service in self.services
            if service.service_type == service_type
        )


@dataclass
class CopernicusMarineProductProvider:
    name: str
    roles: list[str]
    url: str
    email: str


@dataclass
class CopernicusMarineProduct:
    title: str
    product_id: str
    thumbnail_url: str
    description: str
    digital_object_identifier: Optional[str]
    sources: List[str]
    originating_center: str
    processing_level: Optional[str]
    production_center: str
    creation_datetime: Optional[str]
    modified_datetime: Optional[str]
    keywords: dict[str, str]
    datasets: list[CopernicusMarineProductDataset]


@dataclass
class CopernicusMarineCatalogue:
    products: list[CopernicusMarineProduct]

    def filter(self, tokens: list[str]):
        return filter_catalogue_with_strings(self, tokens)


class CatalogParserConnection:
    def __init__(self, proxy: Optional[str] = None) -> None:
        self.proxy = proxy
        self.session = aiohttp.ClientSession(trust_env=True)

    async def get_json_file(self, url: str) -> dict[str, Any]:
        async with self.session.get(
            url,
            params=construct_query_params_for_marine_data_store_monitoring(),
        ) as response:
            return await response.json()

    async def close(self) -> None:
        await self.session.close()


def _construct_copernicus_marine_service(
    stac_service_name, stac_asset
) -> Optional[CopernicusMarineDatasetService]:
    try:
        return CopernicusMarineDatasetService(
            service_type=_service_type_from_web_api_string(stac_service_name),
            uri=stac_asset.get_absolute_href(),
        )
    except ServiceNotHandled as service_not_handled:
        log_exception_debug(service_not_handled)
        return None


def _get_services(
    stac_assets_dict: dict[str, pystac.Asset],
) -> list[CopernicusMarineDatasetService]:
    return [
        dataset_service
        for stac_service_name, stac_asset in stac_assets_dict.items()
        if stac_asset.roles and "data" in stac_asset.roles
        if (
            dataset_service := _construct_copernicus_marine_service(
                stac_service_name, stac_asset
            )
        )
        is not None
    ]


def _get_coordinates(
    dimensions_cube: dict,
) -> dict[str, CopernicusMarineDatasetCoordinates]:
    def _create_coordinate(
        key: str, value: dict
    ) -> CopernicusMarineDatasetCoordinates:
        return CopernicusMarineDatasetCoordinates(
            coordinates_id="depth" if key == "elevation" else key,
            units=value.get("unit") or "",
            minimum_value=value["extent"][0] if "extent" in value else None,
            maximum_value=value["extent"][1] if "extent" in value else None,
            step=value.get("step"),
            values=value.get("values"),
        )

    coordinates_dict = {}
    for key, value in dimensions_cube.items():
        coordinates_dict[key] = _create_coordinate(key, value)
    return coordinates_dict


def _get_variables(
    stac_dataset: pystac.Item,
) -> list[CopernicusMarineDatasetVariable]:
    def _create_variable(
        variable_cube: dict[str, Any],
        bbox: tuple[float, float, float, float],
        coordinates_dict: dict[str, CopernicusMarineDatasetCoordinates],
    ) -> Union[CopernicusMarineDatasetVariable, None]:
        coordinates = variable_cube["dimensions"]
        return CopernicusMarineDatasetVariable(
            short_name=variable_cube["id"],
            standard_name=variable_cube["standardName"],
            units=variable_cube.get("unit") or "",
            bbox=bbox,
            coordinates=[coordinates_dict[key] for key in coordinates],
        )

    coordinates_dict = _get_coordinates(
        stac_dataset.properties["cube:dimensions"]
    )
    bbox = stac_dataset.bbox
    variables: list[Optional[CopernicusMarineDatasetVariable]] = []
    for var_cube in stac_dataset.properties["cube:variables"].values():
        variables += [_create_variable(var_cube, bbox, coordinates_dict)]
    return [var for var in variables if var]


def _construct_marine_data_store_dataset(
    stac_dataset: pystac.Item,
) -> CopernicusMarineProductDataset:
    dataset_id = stac_dataset.id.rsplit("_", maxsplit=1)[
        0
    ]  # Remove the tag e.g.: '_202211'
    return CopernicusMarineProductDataset(
        dataset_id=dataset_id,
        dataset_name=stac_dataset.properties["title"],
        services=_get_services(stac_dataset.get_assets()),
        variables=_get_variables(stac_dataset),
    )


def _construct_marine_data_store_product(
    stac_tuple: Tuple[pystac.Collection, List[pystac.Item]],
) -> CopernicusMarineProduct:
    stac_product, stac_datasets = stac_tuple
    datasets = map_reject_none(
        _construct_marine_data_store_dataset, stac_datasets
    )
    production_center = [
        provider.name
        for provider in stac_product.providers or []
        if "producer" in provider.roles
    ]

    production_center_name = production_center[0] if production_center else ""

    thumbnail = stac_product.assets and stac_product.assets.get("thumbnail")
    digital_object_identifier = (
        stac_product.extra_fields.get("sci:doi", None)
        if stac_product.extra_fields
        else None
    )
    sources = _get_stac_product_property(stac_product, "sources") or []
    originating_center = (
        _get_stac_product_property(stac_product, "originatingCenter") or ""
    )
    processing_level = _get_stac_product_property(
        stac_product, "processingLevel"
    )
    creation_datetime = _get_stac_product_property(
        stac_product, "creationDate"
    )
    modified_datetime = _get_stac_product_property(
        stac_product, "modifiedDate"
    )

    return CopernicusMarineProduct(
        title=stac_product.title or stac_product.id,
        product_id=stac_product.id,
        thumbnail_url=thumbnail.get_absolute_href() if thumbnail else "",
        description=stac_product.description,
        digital_object_identifier=digital_object_identifier,
        sources=sources,
        originating_center=originating_center,
        processing_level=processing_level,
        production_center=production_center_name,
        creation_datetime=creation_datetime or "",
        modified_datetime=modified_datetime or "",
        keywords=stac_product.keywords,
        datasets=sorted(
            [dataset for dataset in datasets],
            key=lambda dataset: dataset.dataset_id,
        ),
    )


def _get_stac_product_property(
    stac_product: pystac.Collection, property_key: str
) -> Optional[Any]:
    properties: Dict[str, str] = (
        stac_product.extra_fields.get("properties", {})
        if stac_product.extra_fields
        else {}
    )
    return properties.get(property_key)


async def async_fetch_items_from_collection(
    connnection: CatalogParserConnection, collection: pystac.Collection
) -> List[pystac.Item]:
    items = []
    for link in collection.get_item_links():
        if not link.owner:
            logger.warning(f"Invalid Item, no owner for: {link.href}")
            continue
        url = (
            MARINE_DATA_STORE_STAC_BASE_URL
            + "/"
            + link.owner.id
            + "/"
            + link.href
        )
        try:
            item_json = await connnection.get_json_file(url)
            items.append(pystac.Item.from_dict(item_json))
        except pystac.STACError as exception:
            message = (
                "Invalid Item: If datetime is None, a start_datetime "
                + "and end_datetime must be supplied."
            )
            if exception.args[0] != message:
                logger.error(exception)
                raise pystac.STACError(exception.args)
    return items


async def async_fetch_collection(
    connection: CatalogParserConnection, url: str
) -> Optional[Tuple[pystac.Collection, List[pystac.Item]]]:
    json_collection = await connection.get_json_file(url)
    try:
        collection = pystac.Collection.from_dict(json_collection)
        items = await async_fetch_items_from_collection(connection, collection)
        return (collection, items)

    except KeyError as exception:
        messages = ["spatial", "temporal"]
        if exception.args[0] not in messages:
            logger.error(exception)
        return None


async def async_fetch_childs(
    connection: CatalogParserConnection, child_links: List[pystac.Link]
) -> Iterator[Optional[Tuple[pystac.Collection, List[pystac.Item]]]]:
    tasks = []
    for link in child_links:
        tasks.append(
            asyncio.ensure_future(
                async_fetch_collection(connection, link.absolute_href)
            )
        )
    return filter(lambda x: x is not None, await asyncio.gather(*tasks))


async def async_fetch_catalog(
    catalog_root_url: str,
    connection: CatalogParserConnection,
    collection_titles_filter: Optional[List[str]] = None,
) -> Tuple[pystac.Catalog, Iterator[pystac.Collection]]:
    json_catalog = await connection.get_json_file(catalog_root_url)
    catalog = pystac.Catalog.from_dict(json_catalog)
    if collection_titles_filter is not None:
        collection_titles_filter_list: List[str] = collection_titles_filter
        child_links = list(
            filter(
                lambda child_link: child_link.title
                not in collection_titles_filter_list,
                catalog.get_child_links(),
            ),
        )
    else:
        child_links = catalog.get_child_links()
    childs = await async_fetch_childs(connection, child_links)
    return catalog, childs


def _retrieve_missing_marine_data_store_products(
    portal_backend: list[CopernicusMarineProduct],
    connection: CatalogParserConnection,
    staging: bool = False,
) -> list[CopernicusMarineProduct]:
    data_store_root_url = (
        MARINE_DATA_STORE_STAC_ROOT_CATALOG_URL
        if not staging
        else MARINE_DATA_STORE_STAC_ROOT_CATALOG_URL_STAGING
    )

    products_from_backend_portal_catalog = [
        product.product_id for product in portal_backend
    ]

    nest_asyncio.apply()
    loop = asyncio.get_event_loop()

    _, marine_data_store_root_collections = loop.run_until_complete(
        async_fetch_catalog(
            catalog_root_url=data_store_root_url,
            connection=connection,
            collection_titles_filter=products_from_backend_portal_catalog,
        )
    )

    products = map_reject_none(
        _construct_marine_data_store_product,
        marine_data_store_root_collections,
    )

    return list(products)


def parse_catalogue(
    overwrite_metadata_cache: bool,
    no_metadata_cache: bool,
    staging: bool = False,
) -> CopernicusMarineCatalogue:
    logger.debug("Parsing catalogue...")
    catalog = _parse_catalogue(
        overwrite_cache=overwrite_metadata_cache,
        ignore_cache=no_metadata_cache,
        _version=version("copernicus-marine-client"),
        staging=staging,
    )
    logger.debug("Catalogue parsed")
    return catalog


@cachier(cache_dir=DEFAULT_CLIENT_BASE_DIRECTORY)
def _parse_catalogue(
    _version: str,  # force cachier to overwrite cache in case of version update
    staging: bool = False,
) -> CopernicusMarineCatalogue:
    progress_bar = tqdm(total=3, desc="Fetching catalog")
    connection = CatalogParserConnection()
    if not staging:
        logger.debug("Parsing portal catalogue...")
        portal_backend_catalog = _parse_portal_backend_catalogue(connection)
        progress_bar.update()
        logger.debug("Portal catalogue parsed")
    else:
        # because of data misalignment we don't want to use data-be endpoint
        portal_backend_catalog = CopernicusMarineCatalogue(products=[])

    marine_data_store_missing_products = (
        _retrieve_missing_marine_data_store_products(
            portal_backend=portal_backend_catalog.products,
            connection=connection,
            staging=staging,
        )
    )
    progress_bar.update()

    full_catalog = CopernicusMarineCatalogue(
        products=sorted(
            portal_backend_catalog.products
            + marine_data_store_missing_products,
            key=lambda product: product.product_id,
        )
    )

    progress_bar.update()
    asyncio.run(connection.close())

    return full_catalog


async def _async_fetch_raw_products(
    product_ids: List[str], connection: CatalogParserConnection
):
    tasks = []
    for product_id in product_ids:
        tasks.append(
            asyncio.ensure_future(
                connection.get_json_file(product_url(product_id))
            )
        )

    return await asyncio.gather(*tasks)


def product_url(product_id: str) -> str:
    base_url = "https://data-be-prd.marine.copernicus.eu/api/dataset"
    return f"{base_url}/{product_id}" + "?variant=detailed-v2"


def variable_title_to_standard_name(variable_title: str) -> str:
    return variable_title.lower().replace(" ", "_")


def variable_to_pick(layer: dict[str, Any]) -> bool:
    return (
        layer["variableId"] != "__DEFAULT__"
        and layer["subsetVariableIds"]
        and len(layer["subsetVariableIds"]) == 1
    )


def _to_service(
    service_name: str, service_url: str
) -> Optional[CopernicusMarineDatasetService]:
    try:
        service_type = _service_type_from_web_api_string(service_name)
        return CopernicusMarineDatasetService(
            service_type=service_type, uri=service_url
        )
    except ServiceNotHandled as service_not_handled:
        log_exception_debug(service_not_handled)
        return None


def to_coordinates(
    subset_attributes: Tuple[str, dict[str, Any]], layer: dict[str, Any]
) -> CopernicusMarineDatasetCoordinates:
    coordinate_name = subset_attributes[0]
    values: Optional[str]
    if coordinate_name == "depth":
        values = layer.get("zValues")
    elif coordinate_name == "time":
        values = layer.get("tValues")
    else:
        values = None
    return CopernicusMarineDatasetCoordinates(
        coordinates_id=subset_attributes[0],
        units=subset_attributes[1]["units"],
        minimum_value=subset_attributes[1]["min"],
        maximum_value=subset_attributes[1]["max"],
        step=subset_attributes[1].get("step"),
        values=values,
    )


def portal_to_service(
    service_uri: Tuple[str, str]
) -> Optional[CopernicusMarineDatasetService]:
    return _to_service(service_uri[0], service_uri[1])


def to_variable(layer: dict[str, Any]) -> CopernicusMarineDatasetVariable:
    return CopernicusMarineDatasetVariable(
        short_name=layer["variableId"],
        standard_name=variable_title_to_standard_name(layer["variableTitle"]),
        units=layer["units"],
        bbox=layer["bbox"],
        coordinates=list(
            map(to_coordinates, layer["subsetAttrs"].items(), repeat(layer))
        ),
    )


def stac_to_service(asset) -> Optional[CopernicusMarineDatasetService]:
    return _to_service(asset[0], asset[1]["href"])


@dataclass
class DistinctDataset:
    dataset_id: str
    layer_elements: Iterable
    raw_services: Dict
    stac_items_values: Optional[Dict]


def to_dataset(
    distinct_dataset: DistinctDataset,
) -> CopernicusMarineProductDataset:
    dataset_id = distinct_dataset.dataset_id
    layer_elements = list(distinct_dataset.layer_elements)
    services_portal = list(
        map_reject_none(
            portal_to_service,
            distinct_dataset.raw_services.items(),
        ),
    )
    services_stac = (
        list(
            map_reject_none(
                stac_to_service,
                distinct_dataset.stac_items_values["assets"].items(),
            ),
        )
        if distinct_dataset.stac_items_values
        else []
    )

    services = list(chain(services_portal, services_stac))

    return CopernicusMarineProductDataset(
        dataset_id=dataset_id,
        dataset_name=layer_elements[0]["subdatasetTitle"],
        services=services,
        variables=list(
            map(to_variable, filter(variable_to_pick, layer_elements))
        ),
    )


def construct_unique_dataset(
    group_layer, raw_services, stac_items
) -> DistinctDataset:
    dataset_id_from_layer = group_layer[0]
    dataset_layer_elements = group_layer[1]
    dataset_raw_services = raw_services[dataset_id_from_layer]

    for stac_dataset_id, stac_items_values in stac_items.items():
        if stac_dataset_id.startswith(dataset_id_from_layer):
            if "--ext--" in stac_dataset_id.split(dataset_id_from_layer)[-1]:
                continue
            else:
                return DistinctDataset(
                    dataset_id=dataset_id_from_layer,
                    layer_elements=dataset_layer_elements,
                    raw_services=dataset_raw_services,
                    stac_items_values=stac_items_values,
                )
    else:
        return DistinctDataset(
            dataset_id=dataset_id_from_layer,
            layer_elements=dataset_layer_elements,
            raw_services=dataset_raw_services,
            stac_items_values=None,
        )


def to_datasets(
    raw_services: dict[str, dict[str, str]],
    layers: dict[str, dict[str, Any]],
    stac_items: dict,
) -> list[CopernicusMarineProductDataset]:
    groups_layers = groupby(
        layers.values(), key=lambda layer: layer["subdatasetId"]
    )
    distinct_datasets = map(
        construct_unique_dataset,
        groups_layers,
        repeat(raw_services),
        repeat(stac_items),
    )

    return sorted(
        map(to_dataset, distinct_datasets),
        key=lambda distinct_dataset: distinct_dataset.dataset_id,
    )


def _parse_product(raw_product: dict[str, Any]) -> CopernicusMarineProduct:
    return CopernicusMarineProduct(
        title=raw_product["title"],
        product_id=raw_product["id"],
        thumbnail_url=raw_product["thumbnailUrl"],
        description=raw_product["abstract"],
        digital_object_identifier=raw_product["doi"],
        sources=raw_product["sources"],
        originating_center=raw_product["originatingCenter"],
        processing_level=raw_product["processingLevel"]
        if "processingLevel" in raw_product
        else None,
        production_center=raw_product["originatingCenter"],
        creation_datetime=raw_product.get("creationDate"),
        modified_datetime=raw_product.get("modifiedDate"),
        keywords=raw_product["keywords"],
        datasets=to_datasets(
            raw_product["services"],
            raw_product["layers"],
            raw_product["stacItems"],
        ),
    )


def _parse_portal_backend_catalogue(
    connection: CatalogParserConnection,
) -> CopernicusMarineCatalogue:
    base_url = "https://data-be-prd.marine.copernicus.eu/api/datasets"
    response = requests.post(
        base_url,
        json={"size": 1000, "includeOmis": True},
    )
    assert response.ok, response.text
    raw_catalogue: dict[str, Any] = loads(response.text)

    nest_asyncio.apply()
    loop = asyncio.get_event_loop()

    results = loop.run_until_complete(
        _async_fetch_raw_products(raw_catalogue["datasets"].keys(), connection)
    )
    return CopernicusMarineCatalogue(
        products=list(map(_parse_product, results))
    )


# ---------------------------------------
# --- Utils function on any catalogue ---
# ---------------------------------------


def get_dataset_from_id(
    catalogue: CopernicusMarineCatalogue, dataset_id: str
) -> CopernicusMarineProductDataset:
    for product in catalogue.products:
        for dataset in product.datasets:
            if dataset_id == dataset.dataset_id:
                return dataset
    error = KeyError(
        f"The requested dataset '{dataset_id}' was not found in the catalogue,"
        " you can use 'copernicus-marine describe --include-datasets "
        "--contains <search_token>' to find datasets"
    )
    raise error


def get_dataset_and_suffix_path_from_url(
    catalogue: CopernicusMarineCatalogue, dataset_url: str
) -> Tuple[CopernicusMarineProductDataset, str]:
    return next_or_raise_exception(
        (
            dataset_and_suffix_url
            for product in catalogue.products
            for dataset in product.datasets
            for service in dataset.services
            if (
                dataset_and_suffix_url := _parse_dataset_and_suffix_url(
                    dataset, service.uri, dataset_url
                )
            )
            is not None
        ),
        KeyError(
            f"The requested dataset URL '{dataset_url}' "
            "was not found in the catalogue, "
            "you can use 'copernicus-marine describe --include-datasets "
            "--contains <search_token>' to find datasets"
        ),
    )


def _parse_dataset_and_suffix_url(
    dataset, service_uri, dataset_url
) -> Optional[Tuple[CopernicusMarineProductDataset, str]]:
    return (
        (dataset, dataset_url.split(service_uri)[1])
        if dataset_url.startswith(service_uri)
        else None
    )


def get_product_from_url(
    catalogue: CopernicusMarineCatalogue, dataset_url: str
) -> CopernicusMarineProduct:
    """
    Return the product object, with its dataset list filtered
    """
    filtered_catalogue = filter_catalogue_with_strings(
        catalogue, [dataset_url]
    )
    if filtered_catalogue is None:
        error = TypeError("filtered catalogue is empty")
        raise error
    if isinstance(filtered_catalogue, CopernicusMarineCatalogue):
        return filtered_catalogue.products[0]
    return filtered_catalogue["products"][0]


def filter_catalogue_with_strings(
    catalogue: CopernicusMarineCatalogue, tokens: list[str]
) -> dict[str, Any]:
    return find_match_object(catalogue, tokens) or {}


def find_match_object(value: Any, tokens: list[str]) -> Any:
    match: Any
    if isinstance(value, str):
        match = find_match_string(value, tokens)
    elif isinstance(value, Enum):
        match = find_match_enum(value, tokens)
    elif isinstance(value, tuple):
        match = find_match_tuple(value, tokens)
    elif isinstance(value, list):
        match = find_match_list(value, tokens)
    elif hasattr(value, "__dict__"):
        match = find_match_dict(value, tokens)
    else:
        match = None
    return match


def find_match_string(string: str, tokens: list[str]) -> Optional[str]:
    return string if any(token in string for token in tokens) else None


def find_match_enum(enum: Enum, tokens: list[str]) -> Any:
    return find_match_object(enum.value, tokens)


def find_match_tuple(tuple: Tuple, tokens: list[str]) -> Optional[list[Any]]:
    return find_match_list(list(tuple), tokens)


def find_match_list(object_list: list[Any], tokens) -> Optional[list[Any]]:
    def find_match(element: Any) -> Optional[Any]:
        return find_match_object(element, tokens)

    filtered_list: list[Any] = list(map_reject_none(find_match, object_list))
    return filtered_list if filtered_list else None


def find_match_dict(
    structure: dict[str, Any], tokens
) -> Optional[dict[str, Any]]:
    filtered_dict = {
        key: find_match_object(value, tokens)
        for key, value in structure.__dict__.items()
        if find_match_object(value, tokens)
    }

    found_match = any(filtered_dict.values())
    if found_match:
        new_dict = dict(structure.__dict__, **filtered_dict)
        structure.__dict__ = new_dict
    return structure if found_match else None
